// this is an auto-generated file. do not edit!

#ifndef VELOCYPACK_VERSION_NUMBER_H
#define VELOCYPACK_VERSION_NUMBER_H 1

#define VELOCYPACK_VERSION "0.1.33"

#define VELOCYPACK_VERSION_MAJOR 0
#define VELOCYPACK_VERSION_MINOR 1
#define VELOCYPACK_VERSION_PATCH 33

#endif
